# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import sys

project = 'ez-a-sync'
copyright = '2024, BobTheBuidler'
author = 'BobTheBuidler'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx.ext.intersphinx',
]

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
}

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']

autodoc_default_options = {
    'private-members': True,
    'special-members': ','.join([
        '__init__',
        '__call__',
        '__getitem__',
        '__iter__',
        '__aiter__',
        '__next__',
        '__anext__',
        '_Done',
        '_AsyncExecutorMixin',
        '_is_protocol',
    ]),
    'inherited-members': True,
    'member-order': 'groupwise',
    # hide private methods that aren't relevant to us here
    'exclude-members': '_abc_impl,_fget,_fset,_fdel,_ASyncSingletonMeta__instances,_ASyncSingletonMeta__lock'
}
autodoc_typehints = "description"
# Don't show class signature with the class' name.
autodoc_class_signature = "separated"

automodule_generate_module_stub = True

sys.path.insert(0, os.path.abspath('./a_sync'))



# mypy: disable-error-code=valid-type
# mypy: disable-error-code=misc
import functools
import logging
from inspect import isawaitable

from a_sync import _helpers, _kwargs
from a_sync._descriptor import ASyncDescriptor
from a_sync._typing import *
from a_sync.modified import ASyncFunction, ASyncFunctionAsyncDefault, ASyncFunctionSyncDefault

if TYPE_CHECKING:
    from a_sync.abstract import ASyncABC

logger = logging.getLogger(__name__)

class ASyncMethodDescriptor(ASyncDescriptor[I, P, T]):
    __wrapped__: AnyFn[P, T]
    async def __call__(self, instance: I, *args: P.args, **kwargs: P.kwargs) -> T:
        # NOTE: This is only used by TaskMapping atm  # TODO: use it elsewhere
        logger.debug("awaiting %s for instance: %s args: %s kwargs: %s", self, instance, args, kwargs)
        return await self.__get__(instance, None)(*args, **kwargs)
    @overload
    def __get__(self, instance: None, owner: Any) -> Self:...
    @overload
    def __get__(self, instance: I, owner: Any) -> "ASyncBoundMethod[I, P, T]":...
    def __get__(self, instance: Optional[I], owner: Any) -> Union[Self, "ASyncBoundMethod[I, P, T]"]:
        if instance is None:
            return self
        try:
            return instance.__dict__[self.field_name]
        except KeyError:
            from a_sync.abstract import ASyncABC
            if self.default == "sync":
                bound = ASyncBoundMethodSyncDefault(instance, self.__wrapped__, **self.modifiers)
            elif self.default == "async":
                bound = ASyncBoundMethodAsyncDefault(instance, self.__wrapped__, **self.modifiers)
            elif isinstance(instance, ASyncABC):
                try:
                    if instance.__a_sync_instance_should_await__:
                        bound = ASyncBoundMethodSyncDefault(instance, self.__wrapped__, **self.modifiers)
                    else:
                        bound = ASyncBoundMethodAsyncDefault(instance, self.__wrapped__, **self.modifiers)
                except AttributeError:
                    bound = ASyncBoundMethod(instance, self.__wrapped__, **self.modifiers)
            else:
                bound = ASyncBoundMethod(instance, self.__wrapped__, **self.modifiers)
            instance.__dict__[self.field_name] = bound
            logger.debug("new bound method: %s", bound)
            return bound
    def __set__(self, instance, value):
        raise RuntimeError(f"cannot set {self.field_name}, {self} is what you get. sorry.")
    def __delete__(self, instance):
        raise RuntimeError(f"cannot delete {self.field_name}, you're stuck with {self} forever. sorry.")

class ASyncMethodDescriptorSyncDefault(ASyncMethodDescriptor[I, P, T]):
    default = "sync"
    any: ASyncFunctionSyncDefault[Concatenate[AnyIterable[I], P], bool]
    all: ASyncFunctionSyncDefault[Concatenate[AnyIterable[I], P], bool]
    min: ASyncFunctionSyncDefault[Concatenate[AnyIterable[I], P], T]
    max: ASyncFunctionSyncDefault[Concatenate[AnyIterable[I], P], T]
    sum: ASyncFunctionSyncDefault[Concatenate[AnyIterable[I], P], T]
    @overload
    def __get__(self, instance: None, owner: Any = None) -> "ASyncMethodDescriptorSyncDefault[I, P, T]":...
    @overload
    def __get__(self, instance: I, owner: Any = None) -> "ASyncBoundMethodSyncDefault[I, P, T]":...
    def __get__(self, instance: Optional[I], owner: Any = None) -> "Union[ASyncMethodDescriptorSyncDefault, ASyncBoundMethodSyncDefault[I, P, T]]":
        if instance is None:
            return self
        try:
            return instance.__dict__[self.field_name]
        except KeyError:
            bound = ASyncBoundMethodSyncDefault(instance, self.__wrapped__, **self.modifiers)
            instance.__dict__[self.field_name] = bound
            logger.debug("new bound method: %s", bound)
            return bound

class ASyncMethodDescriptorAsyncDefault(ASyncMethodDescriptor[I, P, T]):
    default = "async"
    any: ASyncFunctionAsyncDefault[Concatenate[AnyIterable[I], P], bool]
    all: ASyncFunctionAsyncDefault[Concatenate[AnyIterable[I], P], bool]
    min: ASyncFunctionAsyncDefault[Concatenate[AnyIterable[I], P], T]
    max: ASyncFunctionAsyncDefault[Concatenate[AnyIterable[I], P], T]
    sum: ASyncFunctionAsyncDefault[Concatenate[AnyIterable[I], P], T]
    @overload
    def __get__(self, instance: None, owner: Any) -> "ASyncMethodDescriptorAsyncDefault[I, P, T]":...
    @overload
    def __get__(self, instance: I, owner: Any) -> "ASyncBoundMethodAsyncDefault[I, P, T]":...
    def __get__(self, instance: Optional[I], owner: Any) -> "Union[ASyncMethodDescriptorAsyncDefault, ASyncBoundMethodAsyncDefault[I, P, T]]":
        if instance is None:
            return self
        try:
            return instance.__dict__[self.field_name]
        except KeyError:
            bound = ASyncBoundMethodAsyncDefault(instance, self.__wrapped__, **self.modifiers)
            instance.__dict__[self.field_name] = bound
            logger.debug("new bound method: %s", bound)
            return bound

class ASyncBoundMethod(ASyncFunction[P, T], Generic[I, P, T]):
    __slots__ = "__self__",
    def __init__(
        self, 
        instance: I, 
        unbound: AnyFn[Concatenate[I, P], T], 
        **modifiers: Unpack[ModifierKwargs],
    ) -> None:
        self.__self__ = instance
        # First we unwrap the coro_fn and rewrap it so overriding flag kwargs are handled automagically.
        if isinstance(unbound, ASyncFunction):
            modifiers.update(unbound.modifiers)
            unbound = unbound.__wrapped__
        ASyncFunction.__init__(self, unbound, **modifiers)
        functools.update_wrapper(self, unbound)
    def __repr__(self) -> str:
        instance_type = type(self.__self__)
        return f"<{self.__class__.__name__} for function {instance_type.__module__}.{instance_type.__name__}.{self.__name__} bound to {self.__self__}>"
    @overload
    def __call__(self, *args: P.args, sync: Literal[True], **kwargs: P.kwargs) -> T:...
    @overload
    def __call__(self, *args: P.args, sync: Literal[False], **kwargs: P.kwargs) -> Awaitable[T]:...
    @overload
    def __call__(self, *args: P.args, asynchronous: Literal[False], **kwargs: P.kwargs) -> T:...
    @overload
    def __call__(self, *args: P.args, asynchronous: Literal[True], **kwargs: P.kwargs) -> Awaitable[T]:...
    @overload
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> MaybeAwaitable[T]:...
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> MaybeAwaitable[T]:
        logger.debug("calling %s with args: %s kwargs: %s", self, args, kwargs)
        # This could either be a coroutine or a return value from an awaited coroutine,
        #   depending on if an overriding flag kwarg was passed into the function call.
        retval = coro = ASyncFunction.__call__(self, self.__self__, *args, **kwargs)
        if not isawaitable(retval):
            # The coroutine was already awaited due to the use of an overriding flag kwarg.
            # We can return the value.
            pass
        elif self._should_await(kwargs):
            # The awaitable was not awaited, so now we need to check the flag as defined on 'self' and await if appropriate.
            logger.debug("awaiting %s for %s args: %s kwargs: %s", coro, self, args, kwargs)
            retval = _helpers._await(coro)
        logger.debug("returning %s for %s args: %s kwargs: %s", retval, self, args, kwargs)
        return retval  # type: ignore [call-overload, return-value]
    @functools.cached_property
    def __bound_to_a_sync_instance__(self) -> bool:
        from a_sync.abstract import ASyncABC
        return isinstance(self.__self__, ASyncABC)
    @functools.cached_property
    def __is_async_def__(self) -> bool:
        return asyncio.iscoroutinefunction(self.__wrapped__)
    def _should_await(self, kwargs: dict) -> bool:
        if flag := _kwargs.get_flag_name(kwargs):
            return _kwargs.is_sync(flag, kwargs, pop_flag=True)  # type: ignore [arg-type]
        elif self.default:
            return self.default == "sync"
        elif self.__bound_to_a_sync_instance__:
            self.__self__: "ASyncABC"
            return self.__self__.__a_sync_should_await__(kwargs)
        return self.__is_async_def__


class ASyncBoundMethodSyncDefault(ASyncBoundMethod[I, P, T]):
    def __get__(self, instance: Any, owner: Any) -> ASyncFunctionSyncDefault[P, T]:
        return ASyncBoundMethod.__get__(self, instance, owner)
    @overload
    def __call__(self, *args: P.args, sync: Literal[True], **kwargs: P.kwargs) -> T:...
    @overload
    def __call__(self, *args: P.args, sync: Literal[False], **kwargs: P.kwargs) -> Awaitable[T]:...
    @overload
    def __call__(self, *args: P.args, asynchronous: Literal[False], **kwargs: P.kwargs) -> T:...
    @overload
    def __call__(self, *args: P.args, asynchronous: Literal[True], **kwargs: P.kwargs) -> Awaitable[T]:...
    @overload
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:...
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        return ASyncBoundMethod.__call__(self, *args, **kwargs)

class ASyncBoundMethodAsyncDefault(ASyncBoundMethod[I, P, T]):
    def __get__(self, instance: I, owner: Any) -> ASyncFunctionAsyncDefault[P, T]:
        return ASyncBoundMethod.__get__(self, instance, owner)
    @overload
    def __call__(self, *args: P.args, sync: Literal[True], **kwargs: P.kwargs) -> T:...
    @overload
    def __call__(self, *args: P.args, sync: Literal[False], **kwargs: P.kwargs) -> Awaitable[T]:...
    @overload
    def __call__(self, *args: P.args, asynchronous: Literal[False], **kwargs: P.kwargs) -> T:...
    @overload
    def __call__(self, *args: P.args, asynchronous: Literal[True], **kwargs: P.kwargs) -> Awaitable[T]:...
    @overload
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> Awaitable[T]:...
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> Awaitable[T]:
        return ASyncBoundMethod.__call__(self, *args, **kwargs)




import abc
import functools
import logging

from a_sync import _flags, _kwargs, exceptions, modifiers
from a_sync._meta import ASyncMeta
from a_sync._typing import *
from a_sync.exceptions import NoFlagsFound

logger = logging.getLogger(__name__)

class ASyncABC(metaclass=ASyncMeta):

    ##################################
    # Concrete Methods (overridable) #
    ##################################

    def __a_sync_should_await__(self, kwargs: dict) -> bool:
        """Returns a boolean that indicates whether methods of 'instance' should be called as sync or async methods."""
        try:
            # Defer to kwargs always
            return self.__a_sync_should_await_from_kwargs__(kwargs)
        except exceptions.NoFlagsFound:
            # No flag found in kwargs, check for a flag attribute.
            return self.__a_sync_instance_should_await__

    @functools.cached_property
    def __a_sync_instance_should_await__(self) -> bool:
        """
        You can override this if you want. 
        If you want to be able to hotswap instance modes, you can redefine this as a non-cached property.
        """
        return _flags.negate_if_necessary(self.__a_sync_flag_name__, self.__a_sync_flag_value__)
    
    def __a_sync_should_await_from_kwargs__(self, kwargs: dict) -> bool:
        """You can override this if you want."""
        if flag := _kwargs.get_flag_name(kwargs):
            return _kwargs.is_sync(flag, kwargs, pop_flag=True)  # type: ignore [arg-type]
        raise NoFlagsFound("kwargs", kwargs.keys())
    
    @classmethod
    def __a_sync_instance_will_be_sync__(cls, args: tuple, kwargs: dict) -> bool:
        """You can override this if you want."""
        logger.debug("checking `%s.%s.__init__` signature against provided kwargs to determine a_sync mode for the new instance", cls.__module__, cls.__name__)
        if flag := _kwargs.get_flag_name(kwargs):
            sync = _kwargs.is_sync(flag, kwargs)  # type: ignore [arg-type]
            logger.debug("kwargs indicate the new instance created with args %s %s is %ssynchronous", args, kwargs, 'a' if sync is False else '')
            return sync
        logger.debug("No valid flags found in kwargs, checking class definition for defined default")
        return cls.__a_sync_default_mode__()  # type: ignore [arg-type]

    ######################################
    # Concrete Methods (non-overridable) #
    ######################################
    
    @property
    def __a_sync_modifiers__(self: "ASyncABC") -> ModifierKwargs:
        """You should not override this."""
        return modifiers.get_modifiers_from(self)

    ####################
    # Abstract Methods #
    ####################

    @abc.abstractproperty
    def __a_sync_flag_name__(self) -> str:
        pass
    
    @abc.abstractproperty
    def __a_sync_flag_value__(self) -> bool:
        pass
    
    @abc.abstractclassmethod  # type: ignore [arg-type, misc]
    def __a_sync_default_mode__(cls) -> bool:  # type: ignore [empty-body] 
                                                # mypy doesnt recognize this abc member
        pass




import asyncio
import asyncio.futures
import logging

from a_sync._typing import *
from a_sync.primitives.queue import Queue

logger = logging.getLogger(__name__)

async def exhaust_iterator(iterator: AsyncIterator[T], *, queue: Optional[asyncio.Queue] = None) -> None:
    """
    Description:
        Asynchronously iterates over items from the given async iterator and optionally places them into a queue.
    
        This function is a utility to exhaust an async iterator, with an option to forward the iterated items to a provided asyncio.Queue. It's particularly useful when dealing with asynchronous operations that produce items to be consumed by other parts of an application, enabling a producer-consumer pattern.

    Args:
        iterator (AsyncIterator[T]): The async iterator to exhaust.
        queue (Optional[asyncio.Queue]): An optional queue where iterated items will be placed. If None, items are simply consumed.

    Returns:
        None
    """
    async for thing in iterator:
        if queue:
            logger.debug('putting %s from %s to queue %s', thing, iterator, queue)
            queue.put_nowait(thing)
        
async def exhaust_iterators(iterators, *, queue: Optional[asyncio.Queue] = None) -> None:
    """    
    Description:
        Asynchronously iterates over multiple async iterators concurrently and optionally places their items into a queue.
    
        This function leverages asyncio.gather to concurrently exhaust multiple async iterators. It's useful in scenarios where items from multiple async sources need to be processed or collected together, supporting concurrent operations and efficient multitasking.

    Args:
        iterators: A sequence of async iterators to be exhausted concurrently.
        queue (Optional[asyncio.Queue]): An optional queue where items from all iterators will be placed. If None, items are simply consumed.

    Returns:
        None
    """
    await asyncio.gather(*[exhaust_iterator(iterator, queue=queue) for iterator in iterators]) 
    if queue:
        queue.put_nowait(_Done())
    
T0 = TypeVar('T0')
T1 = TypeVar('T1')
T2 = TypeVar('T2')
T3 = TypeVar('T3')
T4 = TypeVar('T4')
T5 = TypeVar('T5')
T6 = TypeVar('T6')
T7 = TypeVar('T7')
T8 = TypeVar('T8')
T9 = TypeVar('T9')

@overload
def as_yielded(*iterators: AsyncIterator[T]) -> AsyncIterator[T]:...
@overload
def as_yielded(iterator0: AsyncIterator[T0], iterator1: AsyncIterator[T1], iterator2: AsyncIterator[T2], iterator3: AsyncIterator[T3], iterator4: AsyncIterator[T4], iterator5: AsyncIterator[T5], iterator6: AsyncIterator[T6], iterator7: AsyncIterator[T7], iterator8: AsyncIterator[T8], iterator9: AsyncIterator[T9]) -> AsyncIterator[Union[T0, T1, T2, T3, T4, T5, T6, T7, T8, T9]]:...
@overload
def as_yielded(iterator0: AsyncIterator[T0], iterator1: AsyncIterator[T1], iterator2: AsyncIterator[T2], iterator3: AsyncIterator[T3], iterator4: AsyncIterator[T4], iterator5: AsyncIterator[T5], iterator6: AsyncIterator[T6], iterator7: AsyncIterator[T7], iterator8: AsyncIterator[T8]) -> AsyncIterator[Union[T0, T1, T2, T3, T4, T5, T6, T7, T8]]:...
@overload
def as_yielded(iterator0: AsyncIterator[T0], iterator1: AsyncIterator[T1], iterator2: AsyncIterator[T2], iterator3: AsyncIterator[T3], iterator4: AsyncIterator[T4], iterator5: AsyncIterator[T5], iterator6: AsyncIterator[T6], iterator7: AsyncIterator[T7]) -> AsyncIterator[Union[T0, T1, T2, T3, T4, T5, T6, T7]]:...
@overload
def as_yielded(iterator0: AsyncIterator[T0], iterator1: AsyncIterator[T1], iterator2: AsyncIterator[T2], iterator3: AsyncIterator[T3], iterator4: AsyncIterator[T4], iterator5: AsyncIterator[T5], iterator6: AsyncIterator[T6]) -> AsyncIterator[Union[T0, T1, T2, T3, T4, T5, T6]]:...
@overload
def as_yielded(iterator0: AsyncIterator[T0], iterator1: AsyncIterator[T1], iterator2: AsyncIterator[T2], iterator3: AsyncIterator[T3], iterator4: AsyncIterator[T4], iterator5: AsyncIterator[T5]) -> AsyncIterator[Union[T0, T1, T2, T3, T4, T5]]:...
@overload
def as_yielded(iterator0: AsyncIterator[T0], iterator1: AsyncIterator[T1], iterator2: AsyncIterator[T2], iterator3: AsyncIterator[T3], iterator4: AsyncIterator[T4]) -> AsyncIterator[Union[T0, T1, T2, T3, T4]]:...
@overload
def as_yielded(iterator0: AsyncIterator[T0], iterator1: AsyncIterator[T1], iterator2: AsyncIterator[T2], iterator3: AsyncIterator[T3]) -> AsyncIterator[Union[T0, T1, T2, T3]]:...
@overload
def as_yielded(iterator0: AsyncIterator[T0], iterator1: AsyncIterator[T1], iterator2: AsyncIterator[T2]) -> AsyncIterator[Union[T0, T1, T2]]:...
@overload
def as_yielded(iterator0: AsyncIterator[T0], iterator1: AsyncIterator[T1]) -> AsyncIterator[Union[T0, T1]]:...
@overload
def as_yielded(iterator0: AsyncIterator[T0], iterator1: AsyncIterator[T1], iterator2: AsyncIterator[T2], *iterators: AsyncIterator[T]) -> AsyncIterator[Union[T0, T1, T2, T]]:...
async def as_yielded(*iterators: AsyncIterator[T]) -> AsyncIterator[T]:  # type: ignore [misc]
    """
    Description:
        Merges multiple async iterators into a single async iterator that yields items as they become available from any of the source iterators.

        This function is designed to streamline the handling of multiple asynchronous data streams by consolidating them into a single asynchronous iteration context. It enables concurrent fetching and processing of items from multiple sources, improving efficiency and simplifying code structure when dealing with asynchronous operations.

        The merging process is facilitated by internally managing a queue where items from the source iterators are placed as they are fetched. This mechanism ensures that the merged stream of items is delivered in an order determined by the availability of items from the source iterators, rather than their original sequence.

    Args:
        *iterators: Variable length list of AsyncIterator objects to be merged.

    Returns:
        AsyncIterator[T]: An async iterator that yields items from the input async iterators as they become available.

    Note:
        This implementation leverages asyncio tasks and queues to efficiently manage the asynchronous iteration and merging process. It handles edge cases such as early termination and exception management, ensuring robustness and reliability.
    """
    queue: Queue[T] = Queue()
    def get_ready() -> List[T]:
        try:
            return queue.get_all_nowait()
        except asyncio.QueueEmpty:
            return []
    task = asyncio.create_task(exhaust_iterators(iterators, queue=queue))
    def done_callback(t: asyncio.Task) -> None:
        if (e := t.exception()) and not next_fut.done(): 
            next_fut.set_exception(e)
    task.add_done_callback(done_callback)
    while not task.done():
        next_fut = asyncio.get_event_loop().create_future()
        get_task = asyncio.create_task(coro=queue.get(), name=str(queue))
        asyncio.futures._chain_future(get_task, next_fut)  # type: ignore [attr-defined]
        for item in (await next_fut, *get_ready()):
            if isinstance(item, _Done):
                task.cancel()
                return
            yield item
            
    if e := task.exception():
        get_task.cancel()
        raise e

class _Done:
    """
    A sentinel class used to signal the completion of processing in the as_yielded function.

    This class acts as a marker to indicate that all items have been processed and the asynchronous iteration can be concluded. It is used internally within the implementation of as_yielded to efficiently manage the termination of the iteration process once all source iterators have been exhausted.
    """
    pass



import asyncio
import functools
import heapq
import logging
import sys

from a_sync._typing import *

logger = logging.getLogger(__name__)

if sys.version_info < (3, 9):
    class _Queue(asyncio.Queue, Generic[T]):
        __slots__ = "_maxsize", "_loop", "_getters", "_putters", "_unfinished_tasks", "_finished"
else:
    class _Queue(asyncio.Queue[T]):
        __slots__ = "_maxsize", "_getters", "_putters", "_unfinished_tasks", "_finished"

class Queue(_Queue[T]):
    # for type hint support, no functional difference
    async def get(self) -> T:
        self._queue
        return await _Queue.get(self)
    def get_nowait(self) -> T:
        return _Queue.get_nowait(self)
    async def put(self, item: T) -> None:
        return _Queue.put(self, item)
    def put_nowait(self, item: T) -> None:
        return _Queue.put_nowait(self, item)
    
    async def get_all(self) -> List[T]:
        """returns 1 or more items"""
        try:
            return self.get_all_nowait()
        except asyncio.QueueEmpty:
            return [await self.get()]
    def get_all_nowait(self) -> List[T]:
        """returns 1 or more items, or raises asyncio.QueueEmpty"""
        values: List[T] = []
        while True:
            try:
                values.append(self.get_nowait())
            except asyncio.QueueEmpty as e:
                if not values:
                    raise asyncio.QueueEmpty from e
                return values
            
    async def get_multi(self, i: int, can_return_less: bool = False) -> List[T]:
        _validate_args(i, can_return_less)
        items = []
        while len(items) < i and not can_return_less:
            try:
                items.extend(self.get_multi_nowait(i - len(items), can_return_less=True))
            except asyncio.QueueEmpty:
                items = [await self.get()]
        return items
    def get_multi_nowait(self, i: int, can_return_less: bool = False) -> List[T]:
        """
        Just like `asyncio.Queue.get_nowait`, but will return `i` items instead of 1.
        Set `can_return_less` to True if you want to receive up to `i` items.
        """
        _validate_args(i, can_return_less)
        items = []
        for _ in range(i):
            try:
                items.append(self.get_nowait())
            except asyncio.QueueEmpty:
                if items and can_return_less:
                    return items
                # put these back in the queue since we didn't return them
                for value in items:
                    self.put_nowait(value)
                raise asyncio.QueueEmpty from None
        return items


class ProcessingQueue(_Queue[Tuple[P, "asyncio.Future[V]"]], Generic[P, V]):
    __slots__ = "func", "num_workers"
    def __init__(
        self, 
        func: Callable[P, Awaitable[V]], 
        num_workers: int, 
        *, 
        return_data: bool = True, 
        name: str = "",
        loop: Optional[asyncio.AbstractEventLoop] = None,
    ) -> None:
        if sys.version_info < (3, 10):
            super().__init__(loop=loop)
        elif loop:
            raise NotImplementedError(f"You cannot pass a value for `loop` in python {sys.version_info}")
        else:
            super().__init__()
        self.func = func
        self.num_workers = num_workers
        self._name = name
        self._no_futs = not return_data
    def __repr__(self) -> str:
        repr_string = f"<{type(self).__name__}"
        if self._name:
            repr_string += f" name={self._name}"
        repr_string += f" func={self.func} num_workers={self.num_workers}"
        if self._unfinished_tasks:
            repr_string += f" pending={self._unfinished_tasks}"
        return f"{repr_string}>"
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> "asyncio.Future[V]":
        return self.put_nowait(*args, **kwargs)
    def __del__(self) -> None:
        if self._unfinished_tasks > 0:
            context = {
                'message': f'{self} was destroyed but has work pending!',
            }
            asyncio.get_event_loop().call_exception_handler(context)
    async def put(self, *args: P.args, **kwargs: P.kwargs) -> "asyncio.Future[V]":
        self._ensure_workers()
        if self._no_futs:
            return await super().put((args, kwargs))
        fut = self._create_future()
        await super().put((args, kwargs, fut))
        return fut
    def put_nowait(self, *args: P.args, **kwargs: P.kwargs) -> "asyncio.Future[V]":
        self._ensure_workers()
        if self._no_futs:
            return super().put_nowait((args, kwargs))
        fut = self._create_future()
        super().put_nowait((args, kwargs, fut))
        return fut
    def _create_future(self) -> "asyncio.Future[V]":
        return asyncio.get_event_loop().create_future()
    def _ensure_workers(self) -> None:
        if self._workers.done():
            raise self._workers.exception()
    @functools.cached_property
    def _workers(self) -> "asyncio.Task[NoReturn]":
        from a_sync.task import create_task
        logger.debug("starting worker task for %s", self)
        task = create_task(
            asyncio.gather(*[
                asyncio.create_task(self._worker_coro(), name=f"{self._name or repr(self)} [Task-{i}]") 
                for i in range(self.num_workers)]), 
            name=f"{self._name or repr(self)} worker Task",
        )
        task._log_destroy_pending = False
        return task
    async def _worker_coro(self) -> NoReturn:
        args: P.args
        kwargs: P.kwargs
        if self._no_futs:
            while True:
                try:
                    args, kwargs = await self.get()
                    await self.func(*args, **kwargs)
                except Exception as e:
                    logger.error("%s in worker for %s!", type(e).__name__, self)
                    logger.exception(e)
                self.task_done()
        else:
            fut: asyncio.Future[V]
            while True:
                try:
                    args, kwargs, fut = await self.get()
                    fut.set_result(await self.func(*args, **kwargs))
                except Exception as e:
                    try:
                        fut.set_exception(e)
                    except UnboundLocalError as u:
                        logger.error("%s for %s is broken!!!", type(self).__name__, self.func)
                        if str(e) != "local variable 'fut' referenced before assignment":
                            logger.exception(u)
                            raise u
                        logger.exception(e)
                        raise e
                self.task_done()


def _validate_args(i: int, can_return_less: bool) -> None:
    if not isinstance(i, int):
        raise TypeError(f"`i` must be an integer greater than 1. You passed {i}")
    if not isinstance(can_return_less, bool):
        raise TypeError(f"`can_return_less` must be boolean. You passed {can_return_less}")
    if i <= 1:
        raise ValueError(f"`i` must be an integer greater than 1. You passed {i}")


_Args = Tuple[Any]
_Kwargs = Tuple[Tuple[str, Any]]
_Key = Tuple[_Args, _Kwargs]

class SmartFuture(asyncio.Future, Generic[T]):
    def __init__(self, queue: "SmartProcessingQueue", key: _Key, *, loop: Optional[asyncio.AbstractEventLoop] = None) -> None:
        super().__init__(loop=loop)
        self._queue = queue
        self._key = key
        self._waiters: Set["asyncio.Task[T]"] = set()
    def __repr__(self):
        return f"<{type(self).__name__} key={self._key} waiters={self.num_waiters} {self._state}>"
    def __await__(self):
        logger.debug("entering %s", self)
        if self.done():
            return self.result()  # May raise too.
        self._asyncio_future_blocking = True
        self._waiters.add(current_task := asyncio.current_task(self._loop))
        logger.debug("awaiting %s", self)
        yield self  # This tells Task to wait for completion.
        self._waiters.remove(current_task)
        if self.num_waiters == 0:
            self._queue._futs.pop(self._key)
        if not self.done():
            raise RuntimeError("await wasn't used with future")
        return self.result()  # May raise too.
    def __lt__(self, other: "SmartFuture") -> bool:
        """heap considers lower values as higher priority so a future with more waiters will be 'less than' a future with less waiters."""
        return self.num_waiters > other.num_waiters
    @property
    def num_waiters(self) -> int:
        return len(self._waiters)


class _PriorityQueueMixin(Generic[T]):
    def _init(self, maxsize):
        self._queue: List[T] = []
    def _put(self, item, heappush=heapq.heappush):
        heappush(self._queue, item)
    def _get(self, heappop=heapq.heappop):
        return heappop(self._queue)

class PriorityProcessingQueue(_PriorityQueueMixin[T], ProcessingQueue[T, V]):
    # NOTE: WIP
    async def put(self, priority: Any, *args: P.args, **kwargs: P.kwargs) -> "asyncio.Future[V]":
        self._ensure_workers()
        fut = asyncio.get_event_loop().create_future()
        await super().put(self, (priority, args, kwargs, fut))
        return fut
    def put_nowait(self, priority: Any, *args: P.args, **kwargs: P.kwargs) -> "asyncio.Future[V]":
        self._ensure_workers()
        fut = self._create_future()
        super().put_nowait(self, (priority, args, kwargs, fut))
        return fut
    def _get(self, heappop=heapq.heappop):
        priority, args, kwargs, fut = heappop(self._queue)
        return args, kwargs, fut

class _VariablePriorityQueueMixin(_PriorityQueueMixin[T]):
    def _get(self, heapify=heapq.heapify, heappop=heapq.heappop):
        "Resort the heap to consider any changes in priorities and pop the smallest value"
        # resort the heap
        heapify(self._queue)
        # take the job with the most waiters
        return heappop(self._queue)
    def _get_key(self, *args, **kwargs) -> _Key:
        return (args, tuple((kwarg, kwargs[kwarg]) for kwarg in sorted(kwargs)))
    def _create_future(self, key: _Key) -> "asyncio.Future[V]":
        return SmartFuture(self, key, loop=asyncio.get_event_loop())

class VariablePriorityQueue(_VariablePriorityQueueMixin[T], asyncio.PriorityQueue):
    """A PriorityQueue subclass that allows priorities to be updated (or computed) on the fly"""
    # NOTE: WIP
             
class SmartProcessingQueue(_VariablePriorityQueueMixin[T], ProcessingQueue[Concatenate[T, P], V]):
    """A PriorityProcessingQueue subclass that will execute jobs with the most waiters first"""
    _no_futs = False
    def __init__(
        self, 
        func: Callable[Concatenate[T, P], Awaitable[V]], 
        num_workers: int, 
        *, 
        name: str = "",
        loop: Optional[asyncio.AbstractEventLoop] = None,
    ) -> None:
        super().__init__(func, num_workers, return_data=True, name=name, loop=loop)
        self._futs: Dict[_Key[T], SmartFuture[T]] = {}
    async def put(self, *args: P.args, **kwargs: P.kwargs) -> SmartFuture[V]:
        self._ensure_workers()
        key = self._get_key(*args, **kwargs)
        if fut := self._futs.get(key, None):
            logger.info("using cached fut")
            return fut
        fut = self._create_future(key)
        self._futs[key] = fut
        await Queue.put(self, (fut, args, kwargs))
        return fut
    def put_nowait(self, *args: P.args, **kwargs: P.kwargs) -> SmartFuture[V]:
        self._ensure_workers()
        key = self._get_key(*args, **kwargs)
        if fut := self._futs.get(key, None):
            logger.info("using cached fut")
            return fut
        fut = self._create_future(key)
        self._futs[key] = fut
        Queue.put_nowait(self, (fut, args, kwargs))
        return fut
    def _get(self):
        fut, args, kwargs = super()._get()
        return args, kwargs, fut
    async def _worker_coro(self) -> NoReturn:
        args: P.args
        kwargs: P.kwargs
        fut: SmartFuture[V]
        while True:
            try:
                args, kwargs, fut = await self.get()
                if fut.num_waiters > 1:
                    logger.info("processing %s", fut)
                else:
                    logger.debug("processing %s", fut)
                result = await self.func(*args, **kwargs)
                fut.set_result(result)
            except Exception as e:
                try:
                    logger.info("%s: %s", type(e).__name__, e)
                    fut.set_exception(e)
                except UnboundLocalError as u:
                    logger.error("%s for %s is broken!!!", type(self).__name__, self.func)
                    if str(e) != "local variable 'fut' referenced before assignment":
                        logger.exception(u)
                        raise u
                    logger.exception(e)
                    raise e
            self.task_done()





import pytest

from a_sync.base import ASyncGenericBase
from a_sync.iter import ASyncIterable, ASyncIterator


async def async_gen():
    yield 0
    yield 1
    yield 2

def test_iterable_wrap():
    assert isinstance(ASyncIterable(async_gen()), ASyncIterable)
    assert isinstance(ASyncIterable.wrap(async_gen()), ASyncIterable)

def test_iterator_wrap():
    assert isinstance(ASyncIterator(async_gen()), ASyncIterator)
    assert isinstance(ASyncIterator.wrap(async_gen()), ASyncIterator)

def test_iterable_sync():
    assert [i for i in ASyncIterable(async_gen())] == [0, 1, 2]
    assert [i for i in ASyncIterable.wrap(async_gen())] == [0, 1, 2]
    assert ASyncIterable.wrap(async_gen()).materialized == [0, 1, 2]

@pytest.mark.asyncio_cooperative
async def test_iterable_async():
    assert [i async for i in ASyncIterable(async_gen())] == [0, 1, 2]
    assert [i async for i in ASyncIterable.wrap(async_gen())] == [0, 1, 2]
    assert await ASyncIterable.wrap(async_gen()) == [0, 1, 2]

def test_iterator_sync():
    iterator = ASyncIterator.wrap(async_gen())
    for i in range(4):
        if i < 3:
            assert next(iterator) == i
        else:
            with pytest.raises(StopIteration):
                next(iterator)
    assert ASyncIterator.wrap(async_gen()).materialized == [0, 1, 2]

@pytest.mark.asyncio_cooperative   
async def test_iterator_async():
    iterator = ASyncIterator.wrap(async_gen())
    for i in range(4):
        if i < 3:
            assert await iterator.__anext__() == i
        else:
            with pytest.raises(StopAsyncIteration):
                await iterator.__anext__()
    assert await ASyncIterator.wrap(async_gen()) == [0, 1, 2]

generator_wrap = ASyncIterator.wrap(async_gen)

def test_generator_sync():
    iterator = generator_wrap()
    for i in range(4):
        if i < 3:
            assert iterator.__next__() == i
        else:
            with pytest.raises(StopIteration):
                iterator.__next__()
    assert generator_wrap().materialized == [0, 1, 2]

@pytest.mark.asyncio_cooperative
async def test_generator_async():
    iterator = generator_wrap()
    for i in range(4):
        if i < 3:
            assert await iterator.__anext__() == i
        else:
            with pytest.raises(StopAsyncIteration):
                await iterator.__anext__()
    assert await generator_wrap() == [0, 1, 2]

class TestGenerator:
    @ASyncIterator.wrap
    async def generate(self):
        yield 0
        yield 1
        yield 2

def test_bound_generator_sync():
    for _ in TestGenerator().generate():
        assert isinstance(_, int)

@pytest.mark.asyncio_cooperative
async def test_bound_generator_async():
    async for _ in TestGenerator().generate():
        assert isinstance(_, int)

class TestGeneratorMeta(ASyncGenericBase):
    async def generate(self):
        yield 0
        yield 1
        yield 2

def test_bound_generator_meta_sync():
    for _ in TestGeneratorMeta().generate():
        assert isinstance(_, int)

@pytest.mark.asyncio_cooperative
async def test_bound_generator_meta_async():
    async for _ in TestGeneratorMeta().generate():
        assert isinstance(_, int)



import asyncio
from concurrent.futures._base import Executor
from decimal import Decimal
from typing import (TYPE_CHECKING, Any, AsyncGenerator, AsyncIterable, AsyncIterator, 
                    Awaitable, Callable, Coroutine, DefaultDict, Deque, Dict, Generator, 
                    Generic, ItemsView, Iterable, Iterator, KeysView, List, Literal,
                    Mapping, NoReturn, Optional, Protocol, Set, Tuple, Type, TypedDict,
                    TypeVar, Union, ValuesView, final, overload, runtime_checkable)

from typing_extensions import Concatenate, ParamSpec, Self, Unpack


T = TypeVar("T")
K = TypeVar("K")
V = TypeVar("V")
I = TypeVar("I")
E = TypeVar('E', bound=Exception)
TYPE = TypeVar("TYPE", bound=Type)
P = ParamSpec("P")

Numeric = Union[int, float, Decimal]

MaybeAwaitable = Union[Awaitable[T], T]

CoroFn = Callable[P, Awaitable[T]]
SyncFn = Callable[P, T]
AnyFn = Union[CoroFn[P, T], SyncFn[P, T]]

class CoroBoundMethod(Protocol[I, P, T]):
    __self__: I
    __call__: Callable[P, Awaitable[T]]
class SyncBoundMethod(Protocol[I, P, T]):
    __self__: I
    __call__: Callable[P, T]
AnyBoundMethod = Union[CoroBoundMethod[Any, P, T], SyncBoundMethod[Any, P, T]]

@runtime_checkable
class AsyncUnboundMethod(Protocol[I, P, T]):
    __get__: Callable[[I, None], CoroBoundMethod[I, P, T]]
@runtime_checkable
class SyncUnboundMethod(Protocol[I, P, T]):
    __get__: Callable[[I, None], SyncBoundMethod[I, P, T]]
AnyUnboundMethod = Union[AsyncUnboundMethod[I, P, T], SyncUnboundMethod[I, P, T]]

AsyncGetterFunction = Callable[[I], Awaitable[T]]
SyncGetterFunction = Callable[[I], T]
AnyGetterFunction = Union[AsyncGetterFunction[I, T], SyncGetterFunction[I, T]]

AsyncDecorator = Callable[[CoroFn[P, T]], CoroFn[P, T]]
AsyncDecoratorOrCoroFn = Union[AsyncDecorator[P, T], CoroFn[P, T]]

DefaultMode = Literal['sync', 'async', None]

CacheType = Literal['memory', None]
SemaphoreSpec = Optional[Union[asyncio.Semaphore, int]]

class ModifierKwargs(TypedDict, total=False):
    default: DefaultMode
    cache_type: CacheType
    cache_typed: bool
    ram_cache_maxsize: Optional[int]
    ram_cache_ttl: Optional[Numeric]
    runs_per_minute: Optional[int]
    semaphore: SemaphoreSpec
    # sync modifiers
    executor: Executor

AnyIterable = Union[AsyncIterable[K], Iterable[K]]
AnyIterableOrAwaitableIterable = Union[AnyIterable[K], Awaitable[AnyIterable[K]]]



import asyncio
import functools
import inspect
import logging

from a_sync import _kwargs
from a_sync._bound import ASyncBoundMethod, ASyncMethodDescriptor, ASyncMethodDescriptorSyncDefault
from a_sync._typing import *
from a_sync import exceptions
from a_sync.base import ASyncGenericBase
from a_sync.iter import ASyncIterator, ASyncGeneratorFunction
from a_sync.modified import ASyncFunction
from a_sync.primitives.queue import Queue, ProcessingQueue
from a_sync.property import _ASyncPropertyDescriptorBase
from a_sync.utils.as_completed import as_completed
from a_sync.utils.gather import Excluder, gather
from a_sync.utils.iterators import as_yielded, exhaust_iterator


logger = logging.getLogger(__name__)

def create_task(coro: Awaitable[T], *, name: Optional[str] = None, skip_gc_until_done: bool = False) -> "asyncio.Task[T]":
    """
    Extends asyncio.create_task to support any Awaitable, manage task lifecycle, and enhance error handling.

    Unlike asyncio.create_task, which requires a coroutine, this function accepts any Awaitable, ensuring broader
    compatibility. It optionally prevents the task from being garbage-collected until completion and provides
    enhanced error management by wrapping exceptions in a custom exception.

    Args:
        coro: An Awaitable object from which to create the task.
        name: Optional name for the task, aiding in debugging.
        skip_gc_until_done: If True, the task is kept alive until it completes, preventing garbage collection.

    Returns:
        An asyncio.Task object created from the provided Awaitable.
    """

    if not asyncio.iscoroutine(coro):
        coro = __await(coro)
    task = asyncio.create_task(coro, name=name)
    if skip_gc_until_done:
        __persist(asyncio.create_task(__persisted_task_exc_wrap(task)))
    return task


MappingFn = Callable[Concatenate[K, P], Awaitable[V]]

class TaskMapping(DefaultDict[K, "asyncio.Task[V]"], AsyncIterable[Tuple[K, V]]):
    """
    A mapping from keys to asyncio Tasks that asynchronously generates and manages tasks based on input iterables.
    """
    
    _destroyed: bool = False
    "Boolean indicating whether his mapping has been consumed and is no longer usable for aggregations."

    _init_loader: Optional["asyncio.Task[None]"] = None
    "An asyncio Task used to preload values from the iterables."

    _init_loader_next: Optional[Callable[[], Awaitable[Tuple[K, V]]]] = None
    "An asyncio Event that indicates the _init_loader started a new task."
    
    __init_loader_coro: Optional[Awaitable[None]] = None
    """An optional asyncio Coroutine to be run by the `_init_loader`"""

    __slots__ = "concurrency", "_wrapped_func", "_wrapped_func_kwargs", "_name", "_next", "__wrapped__", "__dict__"
    def __init__(
        self, 
        wrapped_func: MappingFn[K, P, V] = None, 
        *iterables: AnyIterableOrAwaitableIterable[K], 
        name: str = '', 
        concurrency: Optional[int] = None, 
        **wrapped_func_kwargs: P.kwargs,
    ) -> None:
        """
        Args:
            wrapped_func: A function that takes a key (and optional parameters) and returns an Awaitable.
            *iterables: Any number of iterables whose elements will be used as keys for task generation.
            name: An optional name for the tasks created by this mapping.
            **wrapped_func_kwargs: Keyword arguments that will be passed to `wrapped_func`.
        """

        self.concurrency = concurrency
        "The max number of coroutines that will run at any given time."

        self.__wrapped__ = wrapped_func
        "The original callable used to initialize this mapping without any modifications."""

        wrapped_func = _unwrap(wrapped_func)
        self._wrapped_func = wrapped_func
        "The function used to generate values for each key."

        if isinstance(wrapped_func, ASyncMethodDescriptor):
            if _kwargs.get_flag_name(wrapped_func_kwargs) is None:
                wrapped_func_kwargs["sync"] = False
        self._wrapped_func_kwargs = wrapped_func_kwargs
        "Additional keyword arguments passed to `_wrapped_func`."

        self._name = name
        "Optional name for tasks created by this mapping."

        if iterables:
            self._next = asyncio.Event()
            "An asyncio Event that indicates the next result is ready"
            @functools.wraps(wrapped_func)
            async def _wrapped_set_next(*args: P.args, __a_sync_recursion: int = 0, **kwargs: P.kwargs) -> V:
                try:
                    retval = await wrapped_func(*args, **kwargs)
                    self._next.set()
                    self._next.clear()
                    return retval
                except exceptions.SyncModeInAsyncContextError as e:
                    raise Exception(e, self.__wrapped__)
                except TypeError as e:
                    if recursion > 2 or not (str(e).startswith(wrapped_func.__name__) and "got multiple values for argument" in str(e)):
                        raise e
                    # NOTE: args ordering is clashing with provided kwargs. We can handle this in a hacky way.
                    # TODO: perform this check earlier and pre-prepare the args/kwargs ordering
                    new_args = list(args)
                    new_kwargs = dict(kwargs)
                    try:
                        for i, arg in enumerate(inspect.getfullargspec(self.__wrapped__).args):
                            if arg in kwargs:
                                new_args.insert(i, new_kwargs.pop(arg))
                            else:
                                break
                        return await _wrapped_set_next(*new_args, **new_kwargs, __a_sync_recursion=__a_sync_recursion+1)
                    except TypeError as e2:
                        raise e if str(e2) == "unsupported callable" else e2
            self._wrapped_func = _wrapped_set_next
            init_loader_queue: Queue[Tuple[K, "asyncio.Future[V]"]] = Queue()
            init_loader_coro = exhaust_iterator(self._tasks_for_iterables(*iterables), queue=init_loader_queue)
            try:
                self._init_loader = create_task(init_loader_coro)
            except RuntimeError as e:
                if str(e) != "no running event loop":
                    raise e
                # its okay, we can start it as soon as the loop starts
                self.__init_loader_coro = init_loader_coro
            self._init_loader_next = init_loader_queue.get_all
        else:
            self._init_loader = None
    
    def __repr__(self) -> str:
        return f"<{type(self).__name__} for {self._wrapped_func} ({dict.__repr__(self)}) at {hex(id(self))}>"
    
    def __hash__(self) -> int:
        return id(self)
    
    def __setitem__(self, item: Any, value: Any) -> None:
        raise NotImplementedError("You cannot manually set items in a TaskMapping")
    
    def __getitem__(self, item: K) -> "asyncio.Task[V]":
        try:
            return super().__getitem__(item)
        except KeyError:
            try:
                if self.concurrency:
                    # NOTE: we use a queue instead of a Semaphore to reduce memory use for use cases involving many many tasks
                    fut = self._queue.put_nowait(item)
                else:
                    coro = self._wrapped_func(item, **self._wrapped_func_kwargs)
                    name = f"{self._name}[{item}]" if self._name else f"{item}",
                    fut = create_task(coro=coro, name=name)
                super().__setitem__(item, fut)
                return fut
            except Exception as e:
                raise e from None
        
    def __await__(self) -> Generator[Any, None, Dict[K, V]]:
        """Wait for all tasks to complete and return a dictionary of the results."""
        return self.gather(sync=False).__await__()

    async def __aiter__(self, pop: bool = False) -> AsyncIterator[Tuple[K, V]]:
        """aiterate thru all key-task pairs, yielding the key-result pair as each task completes"""
        yielded = set()
        # if you inited the TaskMapping with some iterators, we will load those
           
        if self._init_loader:
            while not self._init_loader.done():
                await self._init_loader_next()
                while unyielded := [key for key in self if key not in yielded]:
                    if ready := {key: task for key in unyielded if (task:=self[key]).done()}:
                        for key, task in ready.items():
                            self._if_pop_pop(pop, key)
                            yield key, await task
                            yielded.add(key)
                    else:
                        await self._next.wait()
            # loader is already done by this point, but we need to check for exceptions
            await self._init_loader
        else:
            # if you didn't init the TaskMapping with iterators and you didn't start any tasks manually, we should fail
            self._check_empty()
        # if there are any tasks that still need to complete, we need to await them and yield them
        if unyielded := {key: self[key] for key in self if key not in yielded}:
            async for key, value in as_completed(unyielded, aiter=True):
                self._if_pop_pop(pop, key)
                yield key, value

    def __delitem__(self, item: K) -> None:
        task_or_fut = super().__getitem__(item)
        if task_or_fut.done():
            task_or_fut.cancel()
        super().__delitem__(item)

    def keys(self) -> "TaskMappingKeys[K, V]":
        return TaskMappingKeys(super().keys(), self)

    def values(self) -> "TaskMappingValues[K, V]":
        return TaskMappingValues(super().values(), self)
    
    def items(self) -> "TaskMappingValues[K, V]":
        return TaskMappingItems(super().items(), self)
    
    @ASyncGeneratorFunction
    async def map(self, *iterables: AnyIterableOrAwaitableIterable[K], pop: bool = True, yields: Literal['keys', 'both'] = 'both') -> AsyncIterator[Tuple[K, V]]:
        """
            Asynchronously map iterables to tasks and yield their results.

            Args:
            *iterables: Iterables to map over.
            pop: Whether to remove tasks from the internal storage once they are completed.
            yields: Whether to yield 'keys', 'values', or 'both' (key-value pairs).
        
            Yields:
            Depending on `yields`, either keys, values,
            or tuples of key-value pairs representing the results of completed tasks.
        """
        self._check_not_empty()
        async for _ in self._tasks_for_iterables(*iterables):
            async for key, value in self.yield_completed(pop=pop):
                yield _yield(key, value, yields)
        async for key, value in as_completed(self, aiter=True):
            self._if_pop_pop(pop, key)
            yield _yield(key, value, yields)
    
    @ASyncMethodDescriptorSyncDefault
    async def all(self, pop: bool = True) -> bool:
        self._if_pop_check_destroyed(pop)
        async for key, result in self.__aiter__(pop=pop):
            if not bool(result):
                self._if_pop_clear(pop)
                return False
        return True
    
    @ASyncMethodDescriptorSyncDefault
    async def any(self, pop: bool = True) -> bool:
        self._if_pop_check_destroyed(pop)
        async for key, result in self.__aiter__(pop=pop):
            if bool(result):
                self._if_pop_clear(pop)
                return True
        return False
    
    @ASyncMethodDescriptorSyncDefault
    async def max(self, pop: bool = True) -> V:
        self._if_pop_check_destroyed(pop)
        max = None
        async for key, result in self.__aiter__(pop=pop):
            if max is None or result > max:
                max = result
        if min is None:
            raise ValueError("we should not get here")
        return max
    
    @ASyncMethodDescriptorSyncDefault
    async def min(self, pop: bool = True) -> V:
        self._if_pop_check_destroyed(pop)
        min = None
        async for key, result in self.__aiter__(pop=pop):
            if min is None or result < min:
                min = result
        if min is None:
            raise ValueError("we should not get here")
        return min
            
    @ASyncMethodDescriptorSyncDefault
    async def sum(self, pop: bool = False) -> V:
        self._if_pop_check_destroyed(pop)
        retval = None
        async for key, result in self.__aiter__(pop=pop):
            if retval is None:
                retval = 0 + result
            else:
                retval += result
        if retval is None:
            raise ValueError("we should not get here")
        return result

    @ASyncIterator.wrap
    async def yield_completed(self, pop: bool = True) -> AsyncIterator[Tuple[K, V]]:
        """
        Asynchronously yield tuples of key-value pairs representing the results of any completed tasks.

        Args:
            pop: Whether to remove tasks from the internal storage once they are completed.

        Yields:
            Tuples of key-value pairs representing the results of completed tasks.
        """
        for k, task in dict(self).items():
            if task.done():
                self._if_pop_pop(pop, k)
                yield k, await task
    
    @ASyncMethodDescriptorSyncDefault
    async def gather(
        self, 
        return_exceptions: bool = False, 
        exclude_if: Excluder[V] = None,
        tqdm: bool = False,
        **tqdm_kwargs: Any,
    ) -> Dict[K, V]:
        """Wait for all tasks to complete and return a dictionary of the results."""
        if self._init_loader:
            await self._init_loader
        self._check_empty()
        return await gather(self, return_exceptions=return_exceptions, exclude_if=exclude_if, tqdm=tqdm, **tqdm_kwargs)
    
    @overload
    def pop(self, item: K, cancel: bool = False) -> "Union[asyncio.Task[V], asyncio.Future[V]]":...
    @overload
    def pop(self, item: K, default: K, cancel: bool = False) -> "Union[asyncio.Task[V], asyncio.Future[V]]":...
    def pop(self, *args: K, cancel: bool = False) -> "Union[asyncio.Task[V], asyncio.Future[V]]":
        if not cancel:
            return super().pop(*args)
        fut_or_task = super().pop(*args)
        if isinstance(fut_or_task, asyncio.Future) and not fut_or_task.done():
            fut_or_task.cancel()
        return fut_or_task
    
    def clear(self, cancel: bool = False) -> None:
        for k in self:
            self.pop(k, cancel=cancel)

    @functools.cached_property
    def _init_loader(self) -> "asyncio.Task[None]":
        return create_task(self.__init_loader_coro)
    
    @functools.cached_property
    def _queue(self) -> ProcessingQueue:
        fn = functools.partial(self._wrapped_func, **self._wrapped_func_kwargs)
        return ProcessingQueue(fn, self.concurrency)
    
    def _if_pop_check_destroyed(self, pop: bool) -> None:
        if pop:
            if self._destroyed:
                raise RuntimeError(f"{self} has already been consumed")
            self._destroyed = True
    
    def _if_pop_clear(self, pop: bool) -> None:
        if pop:
            self.clear(cancel=True)
    
    @overload
    def _if_pop_pop(self, pop: bool, key: K, cancel: bool = False) -> None:...
    @overload
    def _if_pop_pop(self, pop: bool, key: K, default: T, cancel: bool = False) -> None:...
    def _if_pop_pop(self, pop: bool, *args: K, cancel: bool = False) -> None:
        if pop:
            self.pop(*args, cancel=cancel)
    
    def _check_empty(self) -> None:
        if not self:
            raise exceptions.MappingIsEmptyError
    
    def _check_not_empty(self) -> None:
        if self:
            raise exceptions.MappingNotEmptyError

    @ASyncGeneratorFunction
    async def _tasks_for_iterables(self, *iterables: AnyIterableOrAwaitableIterable[K]) -> AsyncIterator[Tuple[K, "asyncio.Task[V]"]]:
        """Ensure tasks are running for each key in the provided iterables."""
        async for key in as_yielded(*[_yield_keys(iterable) for iterable in iterables]): # type: ignore [attr-defined]
            yield key, self[key]  # ensure task is running


__persisted_tasks: Set["asyncio.Task[Any]"] = set()

async def __await(awaitable: Awaitable[T]) -> T:
    """Wait for the completion of an Awaitable."""
    try:
        return await awaitable
    except RuntimeError as e:
        args = [e, awaitable]
        if isinstance(awaitable, asyncio.tasks._GatheringFuture):
            args.append(awaitable._children)
        raise RuntimeError(*args) from None

    
def __persist(task: "asyncio.Task[Any]") -> None:
    """Add a task to the set of persisted tasks."""
    __persisted_tasks.add(task)
    __prune_persisted_tasks()

def __prune_persisted_tasks():
    """Remove completed tasks from the set of persisted tasks."""
    for task in tuple(__persisted_tasks):
        if task.done():
            if (e := task.exception()) and not isinstance(e, exceptions.PersistedTaskException):
                logger.exception(e)
                raise e
            __persisted_tasks.discard(task)

async def __persisted_task_exc_wrap(task: "asyncio.Task[T]") -> T:
    """
    Wrap a task to handle its exception in a specialized manner.

    Args:
        task: The asyncio Task to wrap.

    Returns:
        The result of the task, if successful.

    Raises:
        PersistedTaskException: Wraps any exception raised by the task for special handling.
    """
    try:
        return await task
    except Exception as e:
        raise exceptions.PersistedTaskException(e, task) from e

@overload
def _yield(key: K, value: V, yields: Literal['keys']) -> K:...
@overload
def _yield(key: K, value: V, yields: Literal['both']) -> Tuple[K, V]:...
def _yield(key: K, value: V, yields: Literal['keys', 'both']) -> Union[K, Tuple[K, V]]:
    """
    Yield either the key, value, or both based on the 'yields' parameter.
    
    Args:
        key: The key of the task.
        value: The result of the task.
        yields: Determines what to yield; 'keys' for keys, 'both' for key-value pairs.
    
    Returns:
        The key, the value, or a tuple of both based on the 'yields' parameter.
    """
    if yields == 'both':
        return key, value
    elif yields == 'keys':
        return key
    else:
        raise ValueError(f"`yields` must be 'keys' or 'both'. You passed {yields}")
    
async def _yield_keys(iterable: AnyIterableOrAwaitableIterable[K]) -> AsyncIterator[K]:
    """
    Asynchronously yield keys from the provided iterable.

    Args:
        iterable: An iterable that can be either synchronous or asynchronous.

    Yields:
        Keys extracted from the iterable.
    """
    if not iterable:
        raise ValueError(iterable)
    elif isinstance(iterable, AsyncIterable):
        async for key in iterable:
            yield key
    elif isinstance(iterable, Iterable):
        for key in iterable:
            yield key
    elif inspect.isawaitable(iterable):
        async for key in _yield_keys(await iterable):
            yield key
    else:
        raise TypeError(iterable)

@functools.lru_cache(maxsize=None)
def _unwrap(wrapped_func: Union[AnyFn[P, T], "ASyncMethodDescriptor[P, T]", _ASyncPropertyDescriptorBase[I, T]]) -> Callable[P, Awaitable[T]]:
    if isinstance(wrapped_func, (ASyncBoundMethod, ASyncMethodDescriptor)):
        return wrapped_func
    elif isinstance(wrapped_func, _ASyncPropertyDescriptorBase):
        return wrapped_func.get
    elif isinstance(wrapped_func, ASyncFunction):
        # this speeds things up a bit by bypassing some logic
        # TODO implement it like this elsewhere if profilers suggest
        return wrapped_func._modified_fn if wrapped_func._async_def else wrapped_func._asyncified
    return wrapped_func


class _TaskMappingView(ASyncGenericBase, Iterable[T], Generic[T, K, V]):
    def __init__(self, view: Iterable[T], task_mapping: TaskMapping[K, V]) -> None:
        self.__view__ = view
        self.__mapping__ = task_mapping
    def __iter__(self) -> Iterator[T]:
        return iter(self.__view__)
    def __await__(self) -> List[T]:
        return self._await().__await__()
    def __len__(self) -> int:
        return len(self.__view__)
    async def _await(self) -> List[T]:
        return [result async for result in self]
    __slots__ = "__view__", "__mapping__"


class TaskMappingKeys(_TaskMappingView[K, K, V], Generic[K, V]):
    async def __aiter__(self) -> AsyncIterator[K]:
        yielded = set()
        for key in self.__mapping__:
            yielded.add(key)
            yield key
        if self.__mapping__._init_loader is None:
            return
        while not self.__mapping__._init_loader.done():
            for key, fut in await self.__mapping__._init_loader_next():
                if key not in yielded:
                    yielded.add(key)
                    yield key
        if e := self.__mapping__._init_loader.exception():
            raise e
        for key in self.__mapping__:
            if key not in yielded:
                yield key

class TaskMappingItems(_TaskMappingView[Tuple[K, V], K, V], Generic[K, V]):
    async def __aiter__(self) -> AsyncIterator[Tuple[K, V]]:
        async for key in self.__mapping__.keys():
            yield key, await self.__mapping__[key]
    
class TaskMappingValues(_TaskMappingView[V, K, V], Generic[K, V]):
    async def __aiter__(self) -> AsyncIterator[V]:
        async for key in self.__mapping__.keys():
            yield await self.__mapping__[key]



import asyncio
import pytest

from a_sync import TaskMapping, create_task, exceptions

@pytest.mark.asyncio_cooperative
async def test_create_task():
    await create_task(coro=asyncio.sleep(0), name='test')

@pytest.mark.asyncio_cooperative
async def test_persistent_task():
    check = False
    async def task():
        await asyncio.sleep(1)
        nonlocal check
        check = True
    create_task(coro=task(), skip_gc_until_done=True)
    # there is no local reference to the newly created task. does it still complete? 
    await asyncio.sleep(2)
    assert check is True

@pytest.mark.asyncio_cooperative
async def test_pruning():
    async def task():
        return
    create_task(coro=task(), skip_gc_until_done=True)
    await asyncio.sleep(0)
    # previously, it failed here
    create_task(coro=task(), skip_gc_until_done=True)

@pytest.mark.asyncio_cooperative
async def test_task_mapping_init():
    tasks = TaskMapping(_coro_fn)
    assert tasks._wrapped_func is _coro_fn, f"{tasks._wrapped_func} , {_coro_fn}, {tasks._wrapped_func == _coro_fn}"
    assert tasks._wrapped_func_kwargs == {}
    assert tasks._name == ""
    tasks = TaskMapping(_coro_fn, name='test', kwarg0=1, kwarg1=None)
    assert tasks._wrapped_func_kwargs == {'kwarg0': 1, 'kwarg1': None}
    assert tasks._name == "test"

@pytest.mark.asyncio_cooperative
async def test_task_mapping():
    tasks = TaskMapping(_coro_fn)
    # does it return the correct type
    assert isinstance(tasks[0], asyncio.Task)
    # does it correctly return existing values
    assert tasks[1] is tasks[1]
    # does the task return the correct value
    assert await tasks[0] == "1"
    # can it do it again
    assert await tasks[0] == "1"
    # can we await the mapping?
    assert await tasks == {0: "1", 1: "22"}
    # can we await one from scratch?
    assert await TaskMapping(_coro_fn, range(5)) == {0: "1", 1: "22", 2: "333", 3: "4444", 4: "55555"}
    assert len(tasks) == 2
    
@pytest.mark.asyncio_cooperative
async def test_task_mapping_map_with_sync_iter():
    tasks = TaskMapping(_coro_fn)
    i = 0
    async for k, v in tasks.map(range(5)):
        assert isinstance(k, int)
        assert isinstance(v, str)
        if i < 4:
            # this shouldn't work since there is a mapping in progress
            with pytest.raises(exceptions.MappingNotEmptyError):
                async for k in tasks.map(range(5)):
                    ...
        i += 1
    tasks = TaskMapping(_coro_fn)
    async for k in tasks.map(range(5), pop=False, yields='keys'):
        assert isinstance(k, int)
    
    # test keys
    for k in tasks.keys():
        assert isinstance(k, int)
    awaited = await tasks.keys()
    assert isinstance(awaited, list)
    for k in awaited:
        assert isinstance(k, int)
    async for k in tasks.keys():
        assert isinstance(k, int)
    
    # test values
    for v in tasks.values():
        assert isinstance(v, asyncio.Future)
        assert isinstance(await v, str)
    awaited = await tasks.values()
    assert isinstance(awaited, list)
    for v in awaited:
        assert isinstance(v, str)
    async for v in tasks.values():
        assert isinstance(v, str)
    
    # test items
    for k, v in tasks.items():
        assert isinstance(k, int)
        assert isinstance(v, asyncio.Future)
        assert isinstance(await v, str)
    awaited = await tasks.items()
    assert isinstance(awaited, list)
    for k, v in awaited:
        assert isinstance(k, int)
        assert isinstance(v, str)
    async for k, v in tasks.items():
        assert isinstance(k, int)
        assert isinstance(v, str)
    
@pytest.mark.asyncio_cooperative
async def test_task_mapping_map_with_async_iter():
    async def async_iter():
        for i in range(5):
            yield i
    tasks = TaskMapping(_coro_fn)
    i = 0
    async for k, v in tasks.map(async_iter()):
        assert isinstance(k, int)
        assert isinstance(v, str)
        if i < 4:
            # this shouldn't work since there is a mapping in progress
            with pytest.raises(exceptions.MappingNotEmptyError):
                async for k in tasks.map(async_iter()):
                    ...
        i += 1
    tasks = TaskMapping(_coro_fn)
    async for k in tasks.map(async_iter(), pop=False, yields='keys'):
        assert isinstance(k, int)
    
    # test keys
    for k in tasks.keys():
        assert isinstance(k, int)
    awaited = await tasks.keys()
    assert isinstance(awaited, list)
    for k in awaited:
        assert isinstance(k, int)
    async for k in tasks.keys():
        assert isinstance(k, int)
    
    # test values
    for v in tasks.values():
        assert isinstance(v, asyncio.Future)
        assert isinstance(await v, str)
    awaited = await tasks.values()
    assert isinstance(awaited, list)
    for v in awaited:
        assert isinstance(v, str)
    async for v in tasks.values():
        assert isinstance(v, str)
    
    # test items
    for k, v in tasks.items():
        assert isinstance(k, int)
        assert isinstance(v, asyncio.Future)
        assert isinstance(await v, str)
    awaited = await tasks.items()
    assert isinstance(awaited, list)
    for k, v in awaited:
        assert isinstance(k, int)
        assert isinstance(v, str)
    async for k, v in tasks.items():
        assert isinstance(k, int)
        assert isinstance(v, str)

def test_taskmapping_views_sync():
    tasks = TaskMapping(_coro_fn, range(5))
    
    # keys are currently empty until the loop has a chance to run
    assert len(tasks.keys()) == 0
    assert len(tasks.values()) == 0
    assert len(tasks.items()) == 0
    
    tasks.gather()

    assert len(tasks.keys()) == 5
    assert len(tasks.values()) == 5
    assert len(tasks.items()) == 5

    for k in tasks.keys():
        assert isinstance(k, int)

    # test values
    for v in tasks.values():
        assert isinstance(v, asyncio.Future)
    
    # test items
    for k, v in tasks.items():
        assert isinstance(k, int)
        assert isinstance(v, asyncio.Future)
        
    assert len(tasks.keys()) == 5
    for k in tasks.keys():
        assert isinstance(k, int)

async def _coro_fn(i: int) -> str:
    i += 1
    return str(i) * i




"""
While not the focus of this lib, this module includes some new primitives and some modified versions of standard asyncio primitives.
"""

from a_sync.primitives.executor import (AsyncProcessPoolExecutor,
                                        AsyncThreadPoolExecutor,
                                        ProcessPoolExecutor,
                                        PruningThreadPoolExecutor,
                                        ThreadPoolExecutor)
from a_sync.primitives.locks import *
from a_sync.primitives.queue import Queue, ProcessingQueue, SmartProcessingQueue

__all__ = [
    "AsyncThreadPoolExecutor",
    "ThreadPoolExecutor",
    "AsyncProcessPoolExecutor",
    "ProcessPoolExecutor",
    "PruningThreadPoolExecutor",
    "Semaphore",
    "ThreadsafeSemaphore",
    "PrioritySemaphore",
    "CounterLock",
    "Event",
    "Queue",
    "ProcessingQueue",
    "SmartProcessingQueue",
]




import functools
import logging

import async_property as ap  # type: ignore [import]

from a_sync import _helpers, config, decorator, exceptions
from a_sync._bound import ASyncBoundMethodAsyncDefault, ASyncMethodDescriptorAsyncDefault
from a_sync._descriptor import ASyncDescriptor
from a_sync._typing import *
from a_sync.modified import ASyncFunction, ASyncFunctionAsyncDefault, ASyncFunctionSyncDefault

if TYPE_CHECKING:
    from a_sync.task import TaskMapping


logger = logging.getLogger(__name__)

class _ASyncPropertyDescriptorBase(ASyncDescriptor[I, Tuple[()], T]):
    any: ASyncFunction[AnyIterable[I], bool]
    all: ASyncFunction[AnyIterable[I], bool]
    min: ASyncFunction[AnyIterable[I], T]
    max: ASyncFunction[AnyIterable[I], T]
    sum: ASyncFunction[AnyIterable[I], T]
    __wrapped__: Callable[[I], T]
    __slots__ = "hidden_method_name", "hidden_method_descriptor", "_fget"
    def __init__(
        self, 
        _fget: AsyncGetterFunction[I, T], 
        field_name: Optional[str] = None,
        **modifiers: config.ModifierKwargs,
    ) -> None:
        super().__init__(_fget, field_name, **modifiers)
        self.hidden_method_name = f"__{self.field_name}__"
        hidden_modifiers = dict(self.modifiers)
        hidden_modifiers["default"] = "async"
        self.hidden_method_descriptor: HiddenMethodDescriptor[T] =  HiddenMethodDescriptor(self.get, self.hidden_method_name, **hidden_modifiers)
        if asyncio.iscoroutinefunction(_fget):
            self._fget = self.__wrapped__
        else:
            self._fget = _helpers._asyncify(self.__wrapped__, self.modifiers.executor)
    @overload
    def __get__(self, instance: None, owner: Any) -> Self:...
    @overload
    def __get__(self, instance: I, owner: Any) -> Awaitable[T]:...
    def __get__(self, instance: Optional[I], owner: Any) -> Union[Self, Awaitable[T]]:
        if instance is None:
            return self
        logger.debug("getting awaitable for %s for instance: %s owner: %s", self, instance, owner)
        awaitable = super().__get__(instance, owner)
        # if the user didn't specify a default behavior, we will defer to the instance
        if _is_a_sync_instance(instance):
            should_await = self.default == "sync" if self.default else instance.__a_sync_instance_should_await__
        else:
            should_await = self.default == "sync" if self.default else not asyncio.get_event_loop().is_running()  
        if should_await:
            logger.debug("awaiting awaitable for %s for instance: %s owner: %s", awaitable, self, instance, owner)
            retval = _helpers._await(awaitable)
        else:
            retval = awaitable
        logger.debug("returning %s for %s for instance: %s owner: %s", retval, self, instance, owner)
        return retval
    async def get(self, instance: I, owner: Any = None) -> T:
        if instance is None:
            raise ValueError(instance)
        logger.debug("awaiting %s for instance %s", self, instance)
        return await super().__get__(instance, owner)
    def map(self, instances: AnyIterable[I], owner: Any = None, concurrency: Optional[int] = None, name: str = "") -> "TaskMapping[I, T]":
        from a_sync.task import TaskMapping
        logger.debug("mapping %s to instances: %s owner: %s", self, instances, owner)
        return TaskMapping(self, instances, owner=owner, name=name or self.field_name, concurrency=concurrency)

class ASyncPropertyDescriptor(_ASyncPropertyDescriptorBase[I, T], ap.base.AsyncPropertyDescriptor):
    pass

class property(ASyncPropertyDescriptor[I, T]):...

class ASyncPropertyDescriptorSyncDefault(property[I, T]):
    default = "sync"
    any: ASyncFunctionSyncDefault[AnyIterable[I], bool]
    all: ASyncFunctionSyncDefault[AnyIterable[I], bool]
    min: ASyncFunctionSyncDefault[AnyIterable[I], T]
    max: ASyncFunctionSyncDefault[AnyIterable[I], T]
    sum: ASyncFunctionSyncDefault[AnyIterable[I], T]
    @overload
    def __get__(self, instance: None, owner: Any) -> Self:...
    @overload
    def __get__(self, instance: I, owner: Any) -> T:...
    def __get__(self, instance: Optional[I], owner: Any) -> Union[Self, T]:
        return _ASyncPropertyDescriptorBase.__get__(self, instance, owner)

class ASyncPropertyDescriptorAsyncDefault(property[I, T]):
    default = "async"
    any: ASyncFunctionAsyncDefault[AnyIterable[I], bool]
    all: ASyncFunctionAsyncDefault[AnyIterable[I], bool]
    min: ASyncFunctionAsyncDefault[AnyIterable[I], T]
    max: ASyncFunctionAsyncDefault[AnyIterable[I], T]
    sum: ASyncFunctionAsyncDefault[AnyIterable[I], T]


ASyncPropertyDecorator = Callable[[AnyGetterFunction[I, T]], property[I, T]]
ASyncPropertyDecoratorSyncDefault = Callable[[AnyGetterFunction[I, T]], ASyncPropertyDescriptorSyncDefault[I, T]]
ASyncPropertyDecoratorAsyncDefault = Callable[[AnyGetterFunction[I, T]], ASyncPropertyDescriptorAsyncDefault[I, T]]

@overload
def a_sync_property(  # type: ignore [misc]
    func: Literal[None] = None,
    **modifiers: Unpack[ModifierKwargs],
) -> ASyncPropertyDecorator[I, T]:...

@overload
def a_sync_property(  # type: ignore [misc]
    func: AnyGetterFunction[I, T],
    **modifiers: Unpack[ModifierKwargs],
) -> ASyncPropertyDescriptor[I, T]:...

@overload
def a_sync_property(  # type: ignore [misc]
    func: Literal[None],
    default: Literal["sync"],
    **modifiers: Unpack[ModifierKwargs],
) -> ASyncPropertyDecoratorSyncDefault[I, T]:...

@overload
def a_sync_property(  # type: ignore [misc]
    func: Literal[None],
    default: Literal["sync"],
    **modifiers: Unpack[ModifierKwargs],
) -> ASyncPropertyDecoratorSyncDefault[I, T]:...

@overload
def a_sync_property(  # type: ignore [misc]
    func: Literal[None],
    default: Literal["async"],
    **modifiers: Unpack[ModifierKwargs],
) -> ASyncPropertyDecoratorAsyncDefault[I, T]:...

@overload
def a_sync_property(  # type: ignore [misc]
    func: Literal[None],
    default: DefaultMode = config.DEFAULT_MODE,
    **modifiers: Unpack[ModifierKwargs],
) -> ASyncPropertyDecorator[I, T]:...
    
@overload
def a_sync_property(  # type: ignore [misc]
    default: Literal["sync"],
    **modifiers: Unpack[ModifierKwargs],
) -> ASyncPropertyDecoratorSyncDefault[I, T]:...
    
@overload
def a_sync_property(  # type: ignore [misc]
    default: Literal["async"],
    **modifiers: Unpack[ModifierKwargs],
) -> ASyncPropertyDecoratorAsyncDefault[I, T]:...
    
@overload
def a_sync_property(  # type: ignore [misc]
    func: AnyGetterFunction[I, T],
    default: Literal["sync"],
    **modifiers: Unpack[ModifierKwargs],
) -> ASyncPropertyDescriptorSyncDefault[I, T]:...
    
@overload
def a_sync_property(  # type: ignore [misc]
    func: AnyGetterFunction[I, T],
    default: Literal["async"],
    **modifiers: Unpack[ModifierKwargs],
) -> ASyncPropertyDescriptorAsyncDefault[I, T]:...
    
@overload
def a_sync_property(  # type: ignore [misc]
    func: AnyGetterFunction[I, T],
    default: DefaultMode = config.DEFAULT_MODE,
    **modifiers: Unpack[ModifierKwargs],
) -> ASyncPropertyDescriptor[I, T]:...
    
def a_sync_property(  # type: ignore [misc]
    func: Union[AnyGetterFunction[I, T], DefaultMode] = None,
    **modifiers: Unpack[ModifierKwargs],
) -> Union[
    ASyncPropertyDescriptor[I, T],
    ASyncPropertyDescriptorSyncDefault[I, T], 
    ASyncPropertyDescriptorAsyncDefault[I, T], 
    ASyncPropertyDecorator[I, T],
    ASyncPropertyDecoratorSyncDefault[I, T],
    ASyncPropertyDecoratorAsyncDefault[I, T],
]:
    func, modifiers = _parse_args(func, modifiers)
    if modifiers.get("default") == "sync":
        descriptor_class = ASyncPropertyDescriptorSyncDefault
    elif modifiers.get("default") == "async":
        descriptor_class = ASyncPropertyDescriptorAsyncDefault
    else:
        descriptor_class = property
    decorator = functools.partial(descriptor_class, **modifiers)
    return decorator if func is None else decorator(func)


class ASyncCachedPropertyDescriptor(_ASyncPropertyDescriptorBase[I, T], ap.cached.AsyncCachedPropertyDescriptor):
    __slots__ = "_fset", "_fdel", "__async_property__"
    def __init__(
        self, 
        _fget: AsyncGetterFunction[I, T], 
        _fset = None, 
        _fdel = None, 
        field_name=None, 
        **modifiers: Unpack[ModifierKwargs],
    ) -> None:
        super().__init__(_fget, field_name, **modifiers)
        self._check_method_sync(_fset, 'setter')
        self._check_method_sync(_fdel, 'deleter')
        self._fset = _fset
        self._fdel = _fdel

class cached_property(ASyncCachedPropertyDescriptor[I, T]):...

class ASyncCachedPropertyDescriptorSyncDefault(cached_property[I, T]):
    """This is a helper class used for type checking. You will not run into any instance of this in prod."""
    @overload
    def __get__(self, instance: None, owner: Any) -> Self:...
    @overload
    def __get__(self, instance: I, owner: Any) -> T:...
    def __get__(self, instance: Optional[I], owner: Any) -> Union[Self, T]:
        return _ASyncPropertyDescriptorBase.__get__(self, instance, owner)


class ASyncCachedPropertyDescriptorAsyncDefault(cached_property[I, T]):
    """This is a helper class used for type checking. You will not run into any instance of this in prod."""

ASyncCachedPropertyDecorator = Callable[[AnyGetterFunction[I, T]], cached_property[I, T]]
ASyncCachedPropertyDecoratorSyncDefault = Callable[[AnyGetterFunction[I, T]], ASyncCachedPropertyDescriptorSyncDefault[I, T]]
ASyncCachedPropertyDecoratorAsyncDefault = Callable[[AnyGetterFunction[I, T]], ASyncCachedPropertyDescriptorAsyncDefault[I, T]]

@overload
def a_sync_cached_property(  # type: ignore [misc]
    func: Literal[None] = None,
    **modifiers: Unpack[ModifierKwargs],
) -> ASyncCachedPropertyDecorator[I, T]:...

@overload
def a_sync_cached_property(  # type: ignore [misc]
    func: AnyGetterFunction[I, T],
    **modifiers: Unpack[ModifierKwargs],
) -> ASyncCachedPropertyDescriptor[I, T]:...

@overload
def a_sync_cached_property(  # type: ignore [misc]
    func: Literal[None],
    default: Literal["sync"],
    **modifiers: Unpack[ModifierKwargs],
) -> ASyncCachedPropertyDecoratorSyncDefault[I, T]:...

@overload
def a_sync_cached_property(  # type: ignore [misc]
    func: Literal[None],
    default: Literal["async"],
    **modifiers: Unpack[ModifierKwargs],
) -> ASyncCachedPropertyDecoratorAsyncDefault[I, T]:...

@overload
def a_sync_cached_property(  # type: ignore [misc]
    func: Literal[None],
    default: DefaultMode,
    **modifiers: Unpack[ModifierKwargs],
) -> ASyncCachedPropertyDecorator[I, T]:...

@overload
def a_sync_cached_property(  # type: ignore [misc]
    default: Literal["sync"],
    **modifiers: Unpack[ModifierKwargs],
) -> ASyncCachedPropertyDecoratorSyncDefault[I, T]:...

@overload
def a_sync_cached_property(  # type: ignore [misc]
    default: Literal["async"],
    **modifiers: Unpack[ModifierKwargs],
) -> ASyncCachedPropertyDecoratorAsyncDefault[I, T]:...
    
@overload
def a_sync_cached_property(  # type: ignore [misc]
    func: AnyGetterFunction[I, T],
    default: Literal["sync"],
    **modifiers: Unpack[ModifierKwargs],
) -> ASyncCachedPropertyDescriptorSyncDefault[I, T]:... 

@overload
def a_sync_cached_property(  # type: ignore [misc]
    func: AnyGetterFunction[I, T],
    default: Literal["async"],
    **modifiers: Unpack[ModifierKwargs],
) -> ASyncCachedPropertyDescriptorAsyncDefault[I, T]:... 

@overload
def a_sync_cached_property(  # type: ignore [misc]
    func: AnyGetterFunction[I, T],
    default: DefaultMode = config.DEFAULT_MODE,
    **modifiers: Unpack[ModifierKwargs],
) -> ASyncCachedPropertyDescriptor[I, T]:...
    
def a_sync_cached_property(  # type: ignore [misc]
    func: Optional[AnyGetterFunction[I, T]] = None,
    **modifiers: Unpack[ModifierKwargs],
) -> Union[
    ASyncCachedPropertyDescriptor[I, T],
    ASyncCachedPropertyDescriptorSyncDefault[I, T], 
    ASyncCachedPropertyDescriptorAsyncDefault[I, T], 
    ASyncCachedPropertyDecorator[I, T],
    ASyncCachedPropertyDecoratorSyncDefault[I, T],
    ASyncCachedPropertyDecoratorAsyncDefault[I, T],
]:
    func, modifiers = _parse_args(func, modifiers)
    if modifiers.get("default") == "sync":
        descriptor_class = ASyncCachedPropertyDescriptorSyncDefault
    elif modifiers.get("default") == "sync":
        descriptor_class = ASyncCachedPropertyDescriptorAsyncDefault
    else:
        descriptor_class = ASyncCachedPropertyDescriptor
    decorator = functools.partial(descriptor_class, **modifiers)
    return decorator if func is None else decorator(func)

class HiddenMethod(ASyncBoundMethodAsyncDefault[I, Tuple[()], T]):
    def __init__(self, instance: I, unbound: AnyFn[Concatenate[I, P], T], field_name: str, **modifiers: _helpers.ModifierKwargs) -> None:
        super().__init__(instance, unbound, **modifiers)
        self.__name__ = field_name
    def __repr__(self) -> str:
        instance_type = type(self.__self__)
        return f"<{self.__class__.__name__} for property {instance_type.__module__}.{instance_type.__name__}.{self.__name__[2:-2]} bound to {self.__self__}>"
    def _should_await(self, kwargs: dict) -> bool:
        try:
            return self.__self__.__a_sync_should_await_from_kwargs__(kwargs)
        except (AttributeError, exceptions.NoFlagsFound):
            return False
    def __await__(self) -> Generator[Any, None, T]:
        return self(sync=False).__await__()

class HiddenMethodDescriptor(ASyncMethodDescriptorAsyncDefault[I, Tuple[()], T]):
    def __get__(self, instance: I, owner: Any) -> HiddenMethod[I, T]:
        if instance is None:
            return self
        try:
            return instance.__dict__[self.field_name]
        except KeyError:
            bound = HiddenMethod(instance, self.__wrapped__, self.field_name, **self.modifiers)
            instance.__dict__[self.field_name] = bound
            logger.debug("new hidden method: %s", bound)
            return bound

def _is_a_sync_instance(instance: object) -> bool:
    try:
        return instance.__is_a_sync_instance__
    except AttributeError:
        from a_sync.abstract import ASyncABC
        is_a_sync = isinstance(instance, ASyncABC)
        instance.__is_a_sync_instance__ = is_a_sync
        return is_a_sync

def _parse_args(func: Union[None, DefaultMode, AsyncGetterFunction[I, T]], modifiers: ModifierKwargs) -> Tuple[Optional[AsyncGetterFunction[I, T]], ModifierKwargs]:
    if func in ['sync', 'async']:
        modifiers['default'] = func
        func = None
    return func, modifiers




import functools

from a_sync import decorator
from a_sync._typing import *
from a_sync.modified import ASyncFunction, ModifiedMixin, ModifierManager

if TYPE_CHECKING:
    from a_sync import TaskMapping

class ASyncDescriptor(ModifiedMixin, Generic[I, P, T]):
    __slots__ = "field_name", "_fget"
    def __init__(
        self, 
        _fget: AnyFn[Concatenate[I, P], T], 
        field_name: Optional[str] = None, 
        **modifiers: ModifierKwargs,
    ) -> None:
        if not callable(_fget):
            raise ValueError(f'Unable to decorate {_fget}')
        self.modifiers = ModifierManager(modifiers)
        if isinstance(_fget, ASyncFunction):
            self.modifiers.update(_fget.modifiers)
            self.__wrapped__ = _fget
        elif asyncio.iscoroutinefunction(_fget):
            self.__wrapped__: AsyncUnboundMethod[I, P, T] = self.modifiers.apply_async_modifiers(_fget)
        else:
            self.__wrapped__ = _fget
        self.field_name = field_name or _fget.__name__
        functools.update_wrapper(self, self.__wrapped__)
    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} for {self.__wrapped__}>"
    def __set_name__(self, owner, name):
        self.field_name = name
    def map(self, *instances: AnyIterable[I], **bound_method_kwargs: P.kwargs) -> "TaskMapping[I, T]":
        from a_sync.task import TaskMapping
        return TaskMapping(self, *instances, **bound_method_kwargs)
    @functools.cached_property
    def all(self) -> ASyncFunction[Concatenate[AnyIterable[I], P], bool]:
        return decorator.a_sync(default=self.default)(self._all)
    @functools.cached_property
    def any(self) -> ASyncFunction[Concatenate[AnyIterable[I], P], bool]:
        return decorator.a_sync(default=self.default)(self._any)
    @functools.cached_property
    def min(self) -> ASyncFunction[Concatenate[AnyIterable[I], P], T]:
        return decorator.a_sync(default=self.default)(self._min)
    @functools.cached_property
    def max(self) -> ASyncFunction[Concatenate[AnyIterable[I], P], T]:
        return decorator.a_sync(default=self.default)(self._max)
    @functools.cached_property
    def sum(self) -> ASyncFunction[Concatenate[AnyIterable[I], P], T]:
        return decorator.a_sync(default=self.default)(self._sum)
    async def _all(self, *instances: AnyIterable[I], concurrency: Optional[int] = None, name: str = "", **kwargs: P.kwargs) -> bool:
        return await self.map(*instances, concurrency=concurrency, name=name, **kwargs).all(pop=True, sync=False)
    async def _any(self, *instances: AnyIterable[I], concurrency: Optional[int] = None, name: str = "", **kwargs: P.kwargs) -> bool:
        return await self.map(*instances, concurrency=concurrency, name=name, **kwargs).any(pop=True, sync=False)
    async def _min(self, *instances: AnyIterable[I], concurrency: Optional[int] = None, name: str = "", **kwargs: P.kwargs) -> T:
        return await self.map(*instances, concurrency=concurrency, name=name, **kwargs).min(pop=True, sync=False)
    async def _max(self, *instances: AnyIterable[I], concurrency: Optional[int] = None, name: str = "", **kwargs: P.kwargs) -> T:
        return await self.map(*instances, concurrency=concurrency, name=name, **kwargs).max(pop=True, sync=False)
    async def _sum(self, *instances: AnyIterable[I], concurrency: Optional[int] = None, name: str = "", **kwargs: P.kwargs) -> T:
        return await self.map(*instances, concurrency=concurrency, name=name, **kwargs).sum(pop=True, sync=False)




import asyncio
import functools
import inspect
import logging

from async_property import async_cached_property

from a_sync import _helpers
from a_sync._typing import *


logger = logging.getLogger(__name__)

class _AwaitableAsyncIterableMixin(AsyncIterable[T]):
    """
    A mixin class defining logic for awaiting an AsyncIterable
    """
    __wrapped__: AsyncIterable[T]
    def __aiter__(self) -> AsyncIterator[T]:
        "Returns an async iterator for the wrapped async iterable."
        return self.__wrapped__.__aiter__()
    def __await__(self) -> Generator[Any, Any, List[T]]:
        """Asynchronously iterates through all contents of ``Self`` and returns a ``list`` containing the results."""
        return self._materialized.__await__()
    @property
    def materialized(self) -> List[T]:
        """Iterates through all contents of ``Self`` and returns a ``list`` containing the results."""
        return _helpers._await(self._materialized)
    @async_cached_property
    async def _materialized(self) -> List[T]:
        """Asynchronously iterates through all contents of ``Self`` and returns a ``list`` containing the results."""
        return [obj async for obj in self]
    __slots__ = '__async_property__', 
    
class ASyncIterable(_AwaitableAsyncIterableMixin[T], Iterable[T]):
    """
    Description:
        A hybrid Iterable/AsyncIterable implementation designed to offer dual compatibility with both synchronous and asynchronous iteration protocols. This class allows objects to be iterated over using either a standard `for` loop or an `async for` loop, making it versatile in scenarios where the mode of iteration (synchronous or asynchronous) needs to be flexible or is determined at runtime.

        The class achieves this by implementing both `__iter__` and `__aiter__` methods, enabling it to return appropriate iterator objects that can handle synchronous and asynchronous iteration, respectively. This dual functionality is particularly useful in codebases that are transitioning between synchronous and asynchronous code, or in libraries that aim to support both synchronous and asynchronous usage patterns without requiring the user to manage different types of iterable objects.

    """
    @classmethod
    def wrap(cls, wrapped: AsyncIterable[T]) -> "ASyncIterable[T]":
        "Class method to wrap an AsyncIterable for backward compatibility."
        logger.warning("ASyncIterable.wrap will be removed soon. Please replace uses with simple instantiation ie `ASyncIterable(wrapped)`")
        return cls(wrapped)
    def __init__(self, async_iterable: AsyncIterable[T]):
        "Initializes the ASyncIterable with an async iterable."
        self.__wrapped__ = async_iterable
        "The wrapped async iterable object."
    def __repr__(self) -> str:
        return f"<{type(self).__name__} for {self.__wrapped__} at {hex(id(self))}>"
    def __iter__(self) -> Iterator[T]:
        "Returns an iterator for the wrapped async iterable."
        yield from ASyncIterator(self.__aiter__())
    __slots__ = "__wrapped__", 

AsyncGenFunc = Callable[P, AsyncGenerator[T, None]]

class ASyncIterator(_AwaitableAsyncIterableMixin[T], Iterator[T]):
    """
    Description:
        A hybrid Iterator/AsyncIterator implementation that bridges the gap between synchronous and asynchronous iteration. This class provides a unified interface for iteration that can seamlessly operate in both synchronous (`for` loop) and asynchronous (`async for` loop) contexts. It allows the wrapping of asynchronous iterable objects or async generator functions, making them usable in synchronous code without explicitly managing event loops or asynchronous context switches.

        By implementing both `__next__` and `__anext__` methods, ASyncIterator enables objects to be iterated using standard iteration protocols while internally managing the complexities of asynchronous iteration. This design simplifies the use of asynchronous iterables in environments or frameworks that are not inherently asynchronous, such as standard synchronous functions or older codebases being gradually migrated to asynchronous IO.

        This class is particularly useful for library developers seeking to provide a consistent iteration interface across synchronous and asynchronous code, reducing the cognitive load on users and promoting code reusability and simplicity.

    """
    def __next__(self) -> T:
        try:
            return asyncio.get_event_loop().run_until_complete(self.__anext__())
        except StopAsyncIteration as e:
            raise StopIteration from e
    @overload
    def wrap(cls, aiterator: AsyncIterator[T]) -> "ASyncIterator[T]":...
    @overload
    def wrap(cls, async_gen_func: AsyncGenFunc[P, T]) -> "ASyncGeneratorFunction[P, T]":...
    @classmethod
    def wrap(cls, wrapped):
        "Class method to wrap either an AsyncIterator or an async generator function."
        if isinstance(wrapped, AsyncIterator):
            logger.warning("This use case for ASyncIterator.wrap will be removed soon. Please replace uses with simple instantiation ie `ASyncIterator(wrapped)`")
            return cls(wrapped)
        elif inspect.isasyncgenfunction(wrapped):
            return ASyncGeneratorFunction(wrapped)
        raise TypeError(f"`wrapped` must be an AsyncIterator or an async generator function. You passed {wrapped}")
    def __init__(self, async_iterator: AsyncIterator[T]):
        "Initializes the ASyncIterator with an async iterator."
        self.__wrapped__ = async_iterator
        "The wrapped async iterator object."
    async def __anext__(self) -> T:
        "Asynchronously fetches the next item from the async iterator."
        return await self.__wrapped__.__anext__()

class ASyncGeneratorFunction(Generic[P, T]):
    """
    Description:
        Encapsulates an asynchronous generator function, providing a mechanism to use it as an asynchronous iterator with enhanced capabilities. This class wraps an async generator function, allowing it to be called with parameters and return an ASyncIterator object. It is particularly useful for situations where an async generator function needs to be used in a manner that is consistent with both synchronous and asynchronous execution contexts.

        The ASyncGeneratorFunction class supports dynamic binding to instances, enabling it to be used as a method on class instances. When accessed as a descriptor, it automatically handles the binding to the instance, thereby allowing the wrapped async generator function to be invoked with instance context ('self') automatically provided. This feature is invaluable for designing classes that need to expose asynchronous generators as part of their interface while maintaining the ease of use and calling semantics similar to regular methods.

        By providing a unified interface to asynchronous generator functions, this class facilitates the creation of APIs that are flexible and easy to use in a wide range of asynchronous programming scenarios. It abstracts away the complexities involved in managing asynchronous generator lifecycles and invocation semantics, making it easier for developers to integrate asynchronous iteration patterns into their applications.
    
    """
    def __init__(self, async_gen_func: AsyncGenFunc[P, T], instance: Any = None) -> None:
        "Initializes the ASyncGeneratorFunction with the given async generator function and optionally an instance."
        self.field_name = async_gen_func.__name__
        "The name of the async generator function."
        self.__wrapped__ = async_gen_func
        "The actual async generator function."
        self.__instance__ = instance
        "The instance the function is bound to, if any."
        functools.update_wrapper(self, self.__wrapped__)
    def __repr__(self) -> str:
        return f"<{type(self).__name__} for {self.__wrapped__} at {hex(id(self))}>"
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> ASyncIterator[T]:
        "Calls the wrapped async generator function with the given arguments and keyword arguments, returning an ASyncIterator."
        if self.__instance__ is None:
            return ASyncIterator(self.__wrapped__(*args, **kwargs))
        return ASyncIterator(self.__wrapped__(self.__instance__, *args, **kwargs))
    def __get__(self, instance: Any, owner: Any) -> "ASyncGeneratorFunction[P, T]":
        "Descriptor method to make the function act like a non-data descriptor."
        if instance is None:
            return self
        try:
            return instance.__dict__[self.field_name]
        except KeyError:
            gen_func = ASyncGeneratorFunction(self.__wrapped__, instance)
            instance.__dict__[self.field_name] = gen_func
            return gen_func



"""Sphinx documentation plugin used to document ASyncFunction instances.

Introduction
============

Usage
-----

The ez-a-sync extension for Sphinx requires Sphinx 2.0 or later.

Add the extension to your :file:`docs/conf.py` configuration module:

.. code-block:: python

    extensions = (...,
                  'a_sync.sphinx.ext')

If you'd like to change the prefix for tasks in reference documentation
then you can change the ``a_sync_function_prefix`` configuration value:

.. code-block:: python

    a_sync_function_prefix = '(function)'  # < default
    a_sync_descriptor_prefix = '(descriptor)'  # < default
    a_sync_generator_function_prefix = '(genfunc)'  # < default

With the extension installed `autodoc` will automatically find
ASyncFunction objects (e.g. when using the automodule directive)
and generate the correct (as well as add a ``(function)`` prefix),
and you can also refer to the tasks using `:task:proj.tasks.add`
syntax.

Use ``.. autotask::`` to alternatively manually document a task.
"""
from inspect import signature

from docutils import nodes
from sphinx.domains.python import PyFunction, PyMethod
from sphinx.ext.autodoc import FunctionDocumenter, MethodDocumenter

from a_sync._descriptor import ASyncDescriptor
from a_sync.iter import ASyncGeneratorFunction
from a_sync.modified import ASyncFunction, ASyncFunctionAsyncDefault, ASyncFunctionSyncDefault



class _ASyncWrapperDocumenter:
    typ: type

    @classmethod
    def can_document_member(cls, member, membername, isattr, parent):
        return isinstance(member, cls.typ) and getattr(member, '__wrapped__') is not None

    def document_members(self, all_members=False):
        pass

    def check_module(self):
        # Normally checks if *self.object* is really defined in the module
        # given by *self.modname*. But since functions decorated with the @task
        # decorator are instances living in the celery.local, we have to check
        # the wrapped function instead.
        wrapped = getattr(self.object, '__wrapped__', None)
        if wrapped and getattr(wrapped, '__module__') == self.modname:
            return True
        return super().check_module()

class _ASyncFunctionDocumenter(_ASyncWrapperDocumenter, FunctionDocumenter):
    def format_args(self):
        wrapped = getattr(self.object, '__wrapped__', None)
        if wrapped is not None:
            sig = signature(wrapped)
            if "self" in sig.parameters or "cls" in sig.parameters:
                sig = sig.replace(parameters=list(sig.parameters.values())[1:])
            return str(sig)
        return ''

class _ASyncMethodDocumenter(_ASyncWrapperDocumenter, MethodDocumenter):
    def format_args(self):
        wrapped = getattr(self.object, '__wrapped__')
        if wrapped is not None:
            return str(signature(wrapped))
        return ''
    
class _ASyncDirective:
    prefix_env: str
    def get_signature_prefix(self, sig):
        return [nodes.Text(getattr(self.env.config, self.prefix_env))]

class _ASyncFunctionDirective(_ASyncDirective, PyFunction):
    pass

class _ASyncMethodDirective(_ASyncDirective, PyMethod):
    pass
    

class ASyncFunctionDocumenter(_ASyncFunctionDocumenter):
    """Document ASyncFunction instance definitions."""
    objtype = 'a_sync_function'
    typ = ASyncFunction
    priority = 15
    #member_order = 11

class ASyncFunctionSyncDocumenter(_ASyncFunctionDocumenter):
    """Document ASyncFunction instance definitions."""
    objtype = 'a_sync_function_sync'
    typ = ASyncFunctionSyncDefault
    priority = 14
    #member_order = 11

class ASyncFunctionAsyncDocumenter(_ASyncFunctionDocumenter):
    """Document ASyncFunction instance definitions."""
    objtype = 'a_sync_function_async'
    typ = ASyncFunctionAsyncDefault
    priority = 13
    #member_order = 11


class ASyncFunctionDirective(_ASyncFunctionDirective):
    prefix_env = "a_sync_function_prefix"

class ASyncFunctionSyncDirective(_ASyncFunctionDirective):
    prefix_env = "a_sync_function_sync_prefix"

class ASyncFunctionAsyncDirective(_ASyncFunctionDirective):
    prefix_env = "a_sync_function_async_prefix"


class ASyncDescriptorDocumenter(_ASyncMethodDocumenter):
    """Document ASyncDescriptor instance definitions."""
    objtype = 'a_sync_descriptor'
    typ = ASyncDescriptor
    #member_order = 11


class ASyncDescriptorDirective(_ASyncMethodDirective):
    """Sphinx task directive."""
    prefix_env = "a_sync_descriptor_prefix"


class ASyncGeneratorFunctionDocumenter(_ASyncFunctionDocumenter):
    """Document ASyncFunction instance definitions."""
    objtype = 'a_sync_generator_function'
    typ = ASyncGeneratorFunction
    #member_order = 11


class ASyncGeneratorFunctionDirective(_ASyncFunctionDirective):
    """Sphinx task directive."""
    prefix_env = "a_sync_generator_function_prefix"

def autodoc_skip_member_handler(app, what, name, obj, skip, options):
    """Handler for autodoc-skip-member event."""
    if isinstance(obj, (ASyncFunction, ASyncDescriptor, ASyncGeneratorFunction)) and getattr(obj, '__wrapped__'):
        if skip:
            return False
    return None


def setup(app):
    """Setup Sphinx extension."""
    app.setup_extension('sphinx.ext.autodoc')
    
    # function
    app.add_autodocumenter(ASyncFunctionDocumenter)
    app.add_autodocumenter(ASyncFunctionSyncDocumenter)
    app.add_autodocumenter(ASyncFunctionAsyncDocumenter)
    app.add_directive_to_domain('py', 'a_sync_function', ASyncFunctionDirective)
    app.add_directive_to_domain('py', 'a_sync_function_sync', ASyncFunctionSyncDirective)
    app.add_directive_to_domain('py', 'a_sync_function_async', ASyncFunctionAsyncDirective)
    app.add_config_value('a_sync_function_sync_prefix', 'ASyncFunction (sync)', True)
    app.add_config_value('a_sync_function_async_prefix', 'ASyncFunction (async)', True)
    app.add_config_value('a_sync_function_prefix', 'ASyncFunction', True)

    # descriptor
    app.add_autodocumenter(ASyncDescriptorDocumenter)
    app.add_directive_to_domain('py', 'a_sync_descriptor', ASyncDescriptorDirective)
    app.add_config_value('a_sync_descriptor_prefix', 'ASyncDescriptor', True)

    # generator
    
    app.add_autodocumenter(ASyncGeneratorFunctionDocumenter)
    app.add_directive_to_domain('py', 'a_sync_generator_function', ASyncGeneratorFunctionDirective)
    app.add_config_value('a_sync_generator_function_prefix', 'ASyncGeneratorFunction', True)

    app.connect('autodoc-skip-member', autodoc_skip_member_handler)

    return {
        'parallel_read_safe': True
    }
